<?php

declare(strict_types=1);

namespace Ramsey\Collection\Test\Mock;

use Ramsey\Collection\AbstractArray;

/**
 * @template T
 * @extends AbstractArray<T>
 */
class MyArray extends AbstractArray
{
}
